#############################################################
###################### Imports ##############################
#############################################################

from os.path import join, isfile, exists, dirname
from os import mkdir
from operator import itemgetter
import time
from datetime import timedelta
from math import ceil,floor,modf

import numpy as np
from scipy.interpolate import interp1d

from epics import caput,caget
from iexcode.instruments.IEX_BL_config import *

from iexcode.instruments.scanRecord import *
from iexcode.instruments.conversions_constants import *
from iexcode.instruments.xrays import energy, scanXAS, BL_energy_tables
from iexcode.instruments.shutters import branch_shutter_close
from iexcode.instruments.VLS_PGM import mono_energy_get
from iexcode.instruments.files_and_folders import get_next_fileNumber
from iexcode.instruments.Logfile import log_print
from iexcode.instruments.Scienta import *

global EA
try:
    print('connecting to Scienta')
    EA = Scienta()
except:
    print("EA is not connected")

###########################################################################
def _EA_init_Motors(Motors):
    """
    Adds Motors to EA object to be used in scanning
    """
    EA.Motors = Motors

def EA_ioc_init(**kwargs):
    """
    run after restarting the 29idcScienta IOC
    kwargs:
        Close_CShutter="Close" default, if something else then it doesn't close the shutter
    """
    kwargs.setdefault("close_shutter",True)
    if kwargs['close_shutter']==True:
        branch_shutter_close('c')
    #Addding HDF5 atttributes
    filepath='/xorApps/epics/synApps_6_1/ioc/29idcScienta/iocBoot/ioc29idcScienta/HDF5Attributes.xml'
    EA._updateAttributes(filepath)
    #Allow EA to make directories
    caput('29idcScienta:HDF1:CreateDirectory',-1)
    #Enabling User CalcOut, needed for BE 
    caput("29idcScienta:userCalcOutEnable.VAL","Enable")
    #Enabling User StringCalc for caQtDM 
    ioc="29idcScienta:"
    pv=ioc+"userStringCalc2"
    caput(pv+".DESC","scan1 trigger2 calc")
    caput(pv+".INAA",ioc+"scan1.T2PV CP NMS")
    caput(pv+".BB",ioc+"HV:ScanTrigger")
    caput(pv+".CALC$","AA==BB")
    caput(ioc+"userStringCalcEnable.VAL","Enable")
    
    #clipping and other processing
    caput("29idcScienta:Proc1:EnableLowClip","Enable")
    caput("29idcScienta:Proc1:LowClip",1)
    caput("29idcScienta:Proc1:EnableHighClip",'Disable')
    caput("29idcScienta:Proc1:EnableOffsetScale","Enable")
    caput("29idcScienta:Proc1:Scale",10)
    caput("29idcScienta:Proc1:Offset",-1)
    
    #Global Detector Setting
    caput('29idcScienta:HV:AngularROImin.VAL',290)
    caput('29idcScienta:HV:AngularROImax.VAL',830) 
    caput('29idcScienta:HV:AngularBinning','Off')   
    
    ### Add here
    #Setting ExpFrames
    caput("29idcScienta:ExpFrames.VAL",1)
    #setting save default lens values
    EA._AcquisitionMode('Spectra')
    time.sleep(0.25)
    try:
        KE = mono_energy_get()
    except:
        KE = 2000
    EA.put(KE, PE=50, LensMode="Angular")
    caput(EA.PHV+"KineticEnergy:TWV.VAL",1)
    caput(EA._Pcam+"Acquire","Acquire")
    
    print("C-Shutter is closed for initialization ")


def folders_EA(userPath,filePrefix="EA",**kwargs):
    """
    For Staff: folder='b', userName='Staff'
    For ARPES: folder ='c'

    setFolders = True; creates the data folder and sets scanRecords/AreaDetectors 
               = False: only creates the data folders; does NOT set 
    """
    kwargs.setdefault('set_folders',True)
    kwargs.setdefault('debug',False)
    try:
        ADplugin=EA._savePlugin 
        dtype=EA.dtype  
        if dtype == "nc":
            df="netCDF"
        else:
            df=EA.dtype  
    except:
        df="h5"

    fpath=join(userPath,df)
    print("\nFolder: " + fpath)

    if exists(fpath):
        if kwargs['debug']==True:
            print("exists")
        try:
            fileNumber=get_next_fileNumber(fpath,filePrefix, debug=False)
        except:
            kwargs['create_only']=True
            print("fileNumber not set, EA not connected, create_only=True")
    else:
        if kwargs['debug']==True:
            print("making")
        mkdir(fpath)
        fileNumber=1
    try:
        if kwargs['set_folders']:
            caput(ADplugin+"FilePath",fpath)
            caput(ADplugin+"FileName",filePrefix)
            caput(ADplugin+"FileNumber",fileNumber)

            #setup AD
            caput(ADplugin+"FileTemplate","%s%s_%4.4d."+dtype)
            caput(ADplugin+"AutoIncrement","Yes")
            caput(ADplugin+"AutoSave","Yes")

        print("EA path: "+fpath)    
        print("Next "+filePrefix+" file: "+str(fileNumber))  
    except:
        df='h5'
        fpath=join(userPath,df)
    
def EA_log_update():
    """
    spectra entries for the log file
    """
    spectra_info = {
        'PE':[EA.PassEnergy,".0f"],
        'lens_mode':[EA.LensMode,"s"],
        "scan_mode":[EA.scan_mode,"s"],
        "KE":[str(EA.KE),"s"],
        "sweeps":[EA.Sweeps,".0f"],
        "frames":[EA.ExpFrames,".0f"]
    }

    try:
        entry_list=[]
        pv_list=[]
        format_list=[]
        for key in spectra_info:
            entry_list.append(key)
            pv_list.append(spectra_info[key][0])
            format_list.append(spectra_info[key][1])
        iex.BL.mda.log.update(entry_list,pv_list,format_list)    
    except:
        print("EAlog did not write to file, check for errors.")

def log_headerEA():##JM - need to update so that we get the keys from log_EA
    s="\nscan    x   y   z   th   chi   phi   T   scan_mode   E1   E2   step   i   f   PE   lens_mode   SES slit #    ID_mode   hv   exit_slit   GRT   TEY1   TEY2   time\n"
    log_print(s)

def EAsweptTime_estimate(EAlist,overhead=[60,.22]):
    """
    estimates the time for spectra with the current analyzer settings
    overhead[0] = intiating scan (seconds)
    overhead[1] = per point (seconds)
    """
    E=1392#image width
    HVscanDIM=2

    B=EAlist[4]*0.000078# PixelEnergy=PassEnergy*0.000078

    C=EAlist[1]# Estart
    A=ceil(EAlist[3]/B)*B # Estep
    D=C+ceil((EAlist[2]-C)/A)*A # Estop

    E0=ceil((D-C)/(ceil(A/B)*B))
    E1=E/min(floor(ceil(A/B)+0.001),E)
    E2=(E/min(floor(ceil(A/B)+0.001),E))%2
    numPnts=E0+E1-E2+1 
    print(numPnts)
    
    PDLY=caget(EA.PHV+"scan"+str(HVscanDIM)+".PDLY")
    DDLY=caget(EA.PHV+"scan"+str(HVscanDIM)+".PDLY")
    time_seconds=overhead[0]+numPnts*(1/17+PDLY+DDLY+overhead[1])  
    print(C,D,B, EAlist[4])
    return str(timedelta(seconds=time_seconds))


def _scanEATrigger(EAlist,before_after,**kwargs):
    """
    before_after="before" sets up scanIOC scanRecord for EA scan and sets prefix to "MDAscan0000"
    before_after="after" clears up scanIOC scanRecord of EA scan and resets prefix to "EA"
        Trigger EA
        Det20 = EA scanNum

        set the EA._savePlugin Prefix to be 'MDAscan0045_'
        **kwargs:
            scan_dim=1
            detTrig=2
    """ 
    kwargs.setdefault("scan_dim",1)
    kwargs.setdefault("detTrig",2)
    kwargs.setdefault("dtype",EA.dtype)
    kwargs.setdefault("detNum",20)
    kwargs.setdefault("prefix","EA")# if not None then over rides the auto
    kwargs.setdefault("debug",False)

    scanPV=iex.BL.ioc+"scan"+str(kwargs["scan_dim"])
    triggerPV=scanPV+".T"+str(kwargs["detTrig"])+"PV"
    
    if kwargs["debug"]:
        print("scanPV: "+scanPV)
        print("before_after: "+before_after)
        
    #setting EA._savePlugin FilePath, FileName,FileNumber
    if before_after == "before":
        _scanEAPrefix("mda",**kwargs)
        scantype, HVscanPV, KElist, parms =EA._spectraSetup(EAlist,**kwargs)
        caput(triggerPV,HVscanPV)
        
    if before_after == "after":
        _scanEAPrefix(kwargs["prefix"],**kwargs)
        caput(triggerPV,"")
        
    if kwargs["debug"]:
        print(triggerPV,caget(triggerPV, as_string=True))
    return
 

def _scanEAPrefix(ptype,**kwargs):
    """
    sets the EA file prefix based on 
    ptype = "mda" -> for "MDAscan"+current MDA file
        else prefix = ptype
    kwargs:
        debug = False (default)
        prefix
    """
    kwargs.setdefault("debug",False)
    kwargs.setdefault("nzeros",4)
    kwargs.setdefault("debug",False)
    
    prefix=""
    if kwargs["debug"]:
        print(ptype)

    if ptype == "mda":
        fpath = iex.BL.mda.filepath()[0:-3]+EA.dtype
        nextMDA = iex.BL.mda.fileNum() 
        prefix = "MDAscan"+str.zfill(str(nextMDA),kwargs["nzeros"])
    else:
        prefix = ptype
    
    if kwargs["debug"]==kwargs["debug"]:
        print("_scanEAPrefix prefix: ",prefix)

    
    #setting the file path for the EA saving
    fpath=caget(EA._savePlugin +"FilePath",as_string=True)
    caput(EA._savePlugin+"FileName",prefix)
    nextNum=get_next_fileNumber(fpath,prefix,**kwargs)
    caput(EA._savePlugin+"FileNumber",nextNum)
    time.sleep(.5)
    if kwargs["debug"]:
        print("FilePath: ",caget(EA._savePlugin +"FilePath", as_string=True))
        print("FileName: ",caget(EA._savePlugin +"FileName", as_string=True))
        print("FileNumber: ",caget(EA._savePlugin +"FileNumber", as_string=True))

def _BE2KE_setupCalc(BE,DESC,CalcOutNum,OutputPV):
    """
    used by scanEA for talking in BE
    """
    pvCalcOut="29idcScienta:userCalcOut"+str(CalcOutNum)
    if len(OutputPV)==0:
        caput(pvCalcOut+"Enable","D")
    else:
        caput(pvCalcOut+"Enable","E")
    caput(pvCalcOut+".DESC", DESC)
    caput(pvCalcOut+".A",BE)
    caput(pvCalcOut+".INPB","29idmono:ENERGY_MON CP NMS")
    caput(pvCalcOut+".INPC","29idcScienta:HV:WorkFunction  CP NMS")
    caput(pvCalcOut+".CALC$","B-A-C")
    caput(pvCalcOut+".OUT",OutputPV+" PP")
    return pvCalcOut
    
def scanEA_reset(**kwargs):
    """resets the IOC after a forced stop
    """
    kwargs.setdefault("scan_dim",1)

    _scanEATrigger([],"after",**kwargs)
    iex.BL.mda.positioners_clear(kwargs["scan_dim"])
    
def scanEA(EAlist,**kwargs):
    """
    Uses the scanRecord (mda) to take multiple Scienta spectra
    EAlist=
        Fixed Mode:["KE"/"BE",CenterEnergy,PassEnergy,Frames,Sweeps] (5)
        Swept Mode:["KE"/"BE",StartEnergy,StopEnergy,StepEnergy,PassEnergy,Frames,Sweeps] (7)
        Baby Sweep (dither):["KE"/"BE",CenterEnergy,PassEnergy,Frames,Sweeps,"BS"] (6)

            (+) BE is positive below Ef
            (-) BE is negative above Ef
            
    **kwargs
        scanIOC=BL_ioc()
        scan_dim=1
        execute: True/False to start the scan => True (default)
        debug=False
        
    """

    kwargs.setdefault('scanIOC',iex.BL.ioc)
    kwargs.setdefault('scan_dim',1)
    kwargs.setdefault('execute',True)
    kwargs.setdefault("debug",False)
    
    if EAlist[0]=="KE" or EAlist[0]=="BE":
        pass
    else:
        print("need to specify BE or KE")

    if EAlist[-1]=='BS':
        sweeps=EAlist[-2]
    else:
        sweeps=EAlist[-1]
        
    if kwargs['debug']:
        print("sweeps: "+str(sweeps))
        
    if EAlist[0]=="BE":
        if len(EAlist)==5: #Fixed
            pvCalcOut1=_BE2KE_setupCalc(EAlist[1],"BE_center",10,"29idcScienta:HV:fixedEnergy.VAL")
            EAlist[1]=caget(pvCalcOut1+'.VAL')
            arrayP1=list(np.full(sweeps,  EAlist[1]))
            iex.BL.mda.fillin_table(pvCalcOut1+'.PROC',"",kwargs["scanIOC"],kwargs["scan_dim"],arrayP1,1)
            EAlist[1]=arrayP1[0]
            if kwargs['debug']:
                print('\npvCalcOut1: ',pvCalcOut1)
                print('Pos1 table:',arrayP1)
        elif len(EAlist)==6: #Baby-Swept
            pvCalcOut1=_BE2KE_setupCalc(EAlist[1],"BE_center",10,"29idcScienta:HV:babySweepCenter.VAL")
            EAlist[1]=caget(pvCalcOut1+'.VAL')
            arrayP1=list(np.full(sweeps, EAlist[1]))
            iex.BL.mda.fillin_table(pvCalcOut1+'PROC',"",kwargs["scanIOC"],kwargs["scan_dim"],arrayP1,1)
            print('\npvCalcOut1: ',pvCalcOut1)
            print('Pos1 table:',arrayP1)
        elif len(EAlist)==7: #Sweep
            pvCalcOut1=_BE2KE_setupCalc(EAlist[1],"BE_start",9,"29idcScienta:HV:sweepStartEnergy.VAL")
            pvCalcOut2=_BE2KE_setupCalc(EAlist[2],"BE_stop",10,"29idcScienta:HV:sweepStopEnergy.VAL")
            EAlist[1]=caget(pvCalcOut1+'.VAL')
            EAlist[2]=caget(pvCalcOut2+'.VAL')
            arrayP1=list(np.full(sweeps, EAlist[1]))
            arrayP2=list(np.full(sweeps, EAlist[2]))
            iex.BL.mda.fillin_table(pvCalcOut1+'.PROC',"",kwargs["scanIOC"],kwargs["scan_dim"],arrayP1,1)
            iex.BL.mda.fillin_table(pvCalcOut2+'.PROC',"",kwargs["scanIOC"],kwargs["scan_dim"],arrayP2,2)
            if kwargs['debug']:
                print("\npvCalcOut1",pvCalcOut1)
                print("\npvCalcOut2",pvCalcOut2)
                print('Pos1 table:',arrayP1)
                print('Pos2 table:',arrayP2)
        EAlist[0]=='KE'
    else:
        _BE2KE_setupCalc(0,"",9,"")
        _BE2KE_setupCalc(0,"",10,"")  
        
    if kwargs['debug']:
        print('/n EAlist => ',EAlist)

    #set up name and add HV trigger and FileNum as det scan1 (sweeps)
    _scanEATrigger(EAlist,"before",**kwargs)
    
    if kwargs['debug']:
        print("Clearing scan positioners and filling in sweeps")  
    #Fill in Sweeps scan
    iex.BL.mda.positioners_clear(**kwargs)
    VAL=""
    RBV=""
    iex.BL.mda.fillin(VAL,RBV,1,sweeps,1,**kwargs)
    if kwargs['debug']:
        scanPV="29id"+kwargs["scanIOC"]+":scan"+str(kwargs["scan_dim"])
        print("scanPV: "+scanPV)     
    
    
        
    #Writing EA parameters
    EAscanType, EAscanPV, KElist, EAparms =EA._spectraSetup(EAlist,**kwargs)    
    
    if kwargs['debug']:
        print("EA._spectraSetup: ")
        print(EAscanType,EAscanPV, KElist)
    
    print(EA._spectraMessage(EAscanType, EAscanPV, KElist))
    #print(EA._EAspectraTime())
    
    #executing the scan
    if kwargs['execute']:
        EA.put(EAparms['KineticEnergy'],EAparms['PassEnergy'],LensMode="Angular",Frames=EAparms['Frames'],**kwargs)
        time.sleep(10)
        EA.put(EAlist[1]-.05,EAlist[-3],LensMode="Angular")
        time.sleep(2)
        iex.BL.mda.go(**kwargs)
        #After scan
        EA_log_update()
        scanEA_reset(**kwargs)
        iex.BL.mda.table_reset_after(**kwargs)
    else:
        return EAparms


        
def scanFM(RoughPositions,thList,EAlist,**kwargs):
    """
    New FermiMap using ScanRecord table scans to move motors
    RoughPositions is a List rough positions from which to interpolate (use RoughPositions_Find())
    thList=[th_start,th_stop,th_step]
    EAlist  to be finish only one scan at the moment[can be a single list if you are only taking a single scan or a list of lists to take multiple scans]
        **kwargs
            scanIOC = BL_ioc()
            scan_dim = 1
        
            logfile(**kwargs)
    """
    kwargs.setdefault("scan_dim",2)
    kwargs.setdefault("debug",False)
    kwargs.setdefault('execute',True)

    # Making Tables and Filling positioners
    x,y,z,th,chi,phi=interpRoughPositions(RoughPositions,thList[0],thList[1],thList[2])
    if kwargs['debug']:
        print(x,y,z,th,chi,phi)
    
    iex.BL.mda.fillin_table(EA.Motors._motor_dictionary["th"][1],EA.Motors._motor_dictionary["th"][0],th,positioner_num=1)
    iex.BL.mda.fillin_table(EA.Motors._motor_dictionary["x"][1],EA.Motors._motor_dictionary["x"][0],x,positioner_num=2)
    iex.BL.mda.fillin_table(EA.Motors._motor_dictionary["y"][1],EA.Motors._motor_dictionary["y"][0],y,positioner_num=3)
    iex.BL.mda.fillin_table(EA.Motors._motor_dictionary["z"][1],EA.Motors._motor_dictionary["z"][0],z,positioner_num=4)

    #setting up EA
    EAkwargs={
    'execute':False,
    'scan_dim':1
    }
    EAparms=scanEA(EAlist,**EAkwargs)
    
    #executing the scan
    if kwargs["execute"]==True:
        print(EAparms)
        EA.put(EAparms['KineticEnergy'],EAparms['PassEnergy'],LensMode="Angular",Frames=EAparms['Frames'],**kwargs)
        iex.BL.mda.go(**kwargs)
        EA_log_update()
        scanEA_reset(**kwargs)


def interpRoughPositions(RoughPositions,thStart,thStop,thStep,**kwargs):
    """
    Interpolate sample position as a function of theta, based on RoughPosition, 
    a list of motor position lists and returns x,y,z,th,chi,phi

    **kwargs:
        kind="cubic" by default, interpolation type ("linear","cubic","quadratic")

    Usage:  
        x,y,z,th,chi,phi=interpRoughPositions(RoughPositions,3,-8,0.5) 

    (direction to minimize backlash)
        RoughPositions=[
                    [x,y,z,th,chi,phi],
                    [x,y,z,th,chi,phi]
                    ]
    """
    kwargs.setdefault('kind','cubic')
    kwargs.setdefault('debug',False)

    #Puts Rough Position in increasing theta order
    RoughPositions=sorted(RoughPositions, key=itemgetter(3))[::-1] 
    RoughPositions=np.array(RoughPositions)
    if kwargs['debug']:
        print('RoughPositions: ',RoughPositions)

    #thlist
    if kwargs['debug']:
        print('ths: ',thStart,thStop,thStep)
    th=np.arange(np.max([thStart,thStop]),np.min([thStart,thStop])-1.0*abs(thStep),-1.0*abs(thStep),dtype=float)
    if kwargs['debug']:
        print('th: ',th)
        
    #interpolating
    def func(th,th_col,m_col,**kwargs):
        f=interp1d(th_col,m_col)
        m=f(th)
        return m
    x=func(th,RoughPositions[:,3], RoughPositions[:,0])
    y=func(th,RoughPositions[:,3], RoughPositions[:,1])
    z=func(th,RoughPositions[:,3], RoughPositions[:,2])
    chi=func(th,RoughPositions[:,3], RoughPositions[:,4])
    phi=func(th,RoughPositions[:,3], RoughPositions[:,5])

    return x,y,z,th,chi,phi

def mvth_interp(RoughPositions, thVal,**kwargs):
    """
    Moves to the interpolated position for a give theta and RoughPosition list
    uses interpRoughPositions
    """
    x,y,z,th,chi,phi=interpRoughPositions(RoughPositions,thVal,thVal-.1,1,**kwargs)
    Pos=[x[0],y[0],z[0],th[0],chi[0],phi[0]]
    #print("Pos = ",Pos)
    EA.Motors.mvsample(Pos)

def scanEA_hv(hv_start_stop_step_lists,EAlist=[],**kwargs):    
    """
    triggers and EAscan for each photon energy in *hvs
    
    hv_start_stop_step_lists, listoflist ... see scanXAS for more info
    
    EAlist = 
        Fixed Mode: EAlist=["KE/BE",CenterEnergy,PassEnergy,Frames,Sweeps] (5)
        Swept Mode: EAlist=["KE/BE",StartEnergy,StopEnergy,StepEnergy,PassEnergy,Frames,Sweeps] (7)
        Baby Sweep (dither): EAlist=["KE"/"BE",CenterEnergy,PassEnergy,Frames,Sweeps,"BS"] (6)
            KE => Auger
            BE => Core levels
            (+) BE is positive below Ef
            (-) BE is negative above Ef
   
    usage:
        scanEA_hv([400,500,1200],EAlist=["BE",-5,200,17*60,1])
        NOTE that EAlist = needs to be written explicitly
        
    **kwargs
        execute: True/False to start the scan => True (default)
        debug: default => False
        QP_ratio: if other than fully periodic

        (Note: scan_dim=2 is hardcoded for photon energy) 
        
    """
    kwargs.setdefault("average",None)
    kwargs.setdefault("execute",True)
    kwargs.setdefault("debug",False)

    if(kwargs['debug']):
        print('EAlist: ',EAlist)

    #setting up EA
    EAkwargs={
    'execute':False,
    'scan_dim':1
    }
    EAparms=scanEA(EAlist,**EAkwargs)

    scanGo = kwargs['execute']
    #Setting up the ScanRecord for Mono and ID in Table mode
    kwargs["scan_dim"]=2
    kwargs["execute"]=False
    mono_array,ID_array = BL_energy_tables(hv_start_stop_step_lists,**kwargs)
    kwargs.update({"ID_tracking":True})
    scanXAS(hv_start_stop_step_lists,**kwargs)
   
    if scanGo == True:
        #Setting the beamline energy to the first point, and EA at first KE
        energy(mono_array[0])
        #Scanning            
        EA.put(mono_array[0]-EAlist[1],EAparms['PassEnergy'],LensMode="Angular",Frames=EAparms['Frames'],**kwargs)
        iex.BL.mda.go(**kwargs)
        #After scan
        scanEA_reset(**kwargs)
        iex.BL.mda.table_reset_after(**kwargs)

def scanEA_y(EAlist, start,stop,step,mode='absolute',**kwargs):
    scanEA_motor(EAlist,'y',start,stop,step,mode=mode,**kwargs)

def scanEA_z(EAlist, start,stop,step,mode='absolute',**kwargs):
    scanEA_motor(EAlist,'z',start,stop,step,mode=mode,**kwargs)

    
def scanEA_motor(EAlist, motor,start,stop,step,mode='absolute',**kwargs):
    """
    scans an ARPES motor (scan_dim=2) while triggering the EA (scan_dim=1)
       
    EAlist = 
        Fixed Mode:["BE",CenterEnergy,PassEnergy,Frames,Sweeps] (5)
        Swept Mode:["BE",StartEnergy,StopEnergy,StepEnergy,PassEnergy,Frames,Sweeps] (7)
        Baby Sweep (dither):["KE"/"BE",CenterEnergy,PassEnergy,Frames,Sweeps,"BS"] (6)

            (+) BE is positive below Ef
            (-) BE is negative above Ef
        
    **kwargs
        execute: True/False to start the scan => True (default)
        debug=False
        (Note: scan_dim=2 is hardcoded) 
   """
    kwargs.setdefault("execute",True)
    kwargs.setdefault("debug",False)
    
    #setting up EA
    EAkwargs={
    'execute':False,
    'scan_dim':1
    }
    EAparms=scanEA(EAlist,**EAkwargs)
    
    #Setting up the ScanRecord for motor scans
    scan_dim=2 #hard coded
    kwargs.update({'scan_dim':scan_dim})

    EA.Motors.scan(motor,start,stop,step,**kwargs)
    
    if kwargs['debug']:
        print("ScanGo scan_dim = ",scan_dim)
    
    if kwargs["execute"]==True:
        #Scanning            
        EA.put(EAlist[1],EAparms['PassEnergy'],LensMode="Angular",Frames=EAparms['Frames'],**kwargs)
        iex.BL.mda.go(kwargs["scanIOC"],scan_dim)
        EA_log_update()
        #After scan
        scanEA_reset(**kwargs)

def scanEA_Mesh(EAlist,y_start_stop_step,z_start_stop_step,**kwargs):
    """
    2D scan mesh (y/z) while triggering the EA: scan_dim=1
    y_start_stop_step=[start1,stop1,step1] #y: scan_dim=2
    z_start_stop_step=[start2,stop2,step2] #z: scan_dim=3
    
       
    EAlist = 
        Fixed Mode:["BE",CenterEnergy,PassEnergy,Frames,Sweeps] (5)
        Swept Mode:["BE",StartEnergy,StopEnergy,StepEnergy,PassEnergy,Frames,Sweeps] (7)
        Baby Sweep (dither):["KE"/"BE",CenterEnergy,PassEnergy,Frames,Sweeps,"BS"] (6)
1205
            (+) BE is positive below Ef
            (-) BE is negative above Ef
        
    **kwargs
        execute: True/False to start the scan => True (default)
        debug=False
        (Note: scan_dims are hardcoded) 
   """
    kwargs.setdefault("execute",True)
    kwargs.setdefault("debug",False)
    
    #setting up EA
    EAkwargs={
    'execute':False,
    'scan_dim':1
    }
    EAparms=scanEA(EAlist,**EAkwargs)

    #Setting up the ScanRecord for motor scans
    outer_scan_dim=3 #hard coded 
    inner_loop_list = y_start_stop_step.insert(0,"y")
    outer_loop_list = z_start_stop_step.insert(0,"z")
    EA.Motors.scan_2D(inner_loop_list,outer_loop_list,outer_scan_dim,**kwargs)
    
    if kwargs['debug']:
        print("ScanGo scan_dim = ",outer_scan_dim)

    if kwargs['execute']:
        #Scanning            
        EA.put(EAlist[1],EAparms['PassEnergy'],LensMode="Angular",Frames=EAparms['Frames'],**kwargs)
        iex.BL.mda.go(scan_dim=outer_scan_dim)
        EA_log_update()

        #After scan
        scanEA_reset(**kwargs)

        
        
        
    
def hv2kz(lattice,V0,hv):
    """
    Converts a hv value for the nth zone had returns corresponding kz values 
    [0]for zone  boundary and [1] for zone center in inverse angstroms
    lattice = c; assuming kz orthoganal to a-b plane (i.e. 2np.pi/c = GZ distance)
    and at zone center (kx,ky)=0 /cos(th=0)=1
    V0 = the inner potential
    """
    work_fct=EA.wk
    Ek=hv-work_fct
    k_z=0.5124*np.sqrt(Ek+V0)    # Kz at (kx,ky)=0 i.e. cos(th)=1
    c_star=2*np.pi/lattice    # 2pi/c = GZ distance
    GZ_n=round((k_z/c_star*2),2)
    G_n=round((k_z/c_star),2)
    print("  kz = "+str(round(k_z,2))+" A^(-1) = " +str(GZ_n)+" * pi/c = " +str(G_n)+" * 2pi/c")
    return GZ_n,G_n
    
def kz2hv(lattice,V0,n):
    """
    Converts a kz value for the nth zone had returns corresponding hv
    lattice = c; assuming kz orthoganal to a-b plane (i.e. 2pi/c = GZ distance)
    and at zone center (kx,ky)=0 /cos(th=0)=1
    V0 = the inner potential
    """
    work_fct=EA.wk
    c_star=2*np.pi/lattice        # 2pi/c = GZ distance
    Ek=(n*c_star/0.5124)**2-V0    # Ek at (kx,ky)=0 i.e. cos(th)=1
    hv=Ek+work_fct
    mono=round(hv,1)
    print("\n")
    print("  hv = Ek + Phi = "+str(round(hv,2))+" eV")
    print("  kz = n*2pi/c   with  n = "+str(n))
    return mono


def Print_Gamma_n(lattice,V0,n1,n2):
    work_fct=EA.wk
    c_star=2*np.pi/lattice                # 2pi/c = GZ distance
    for n in range(n1,n2+1,1):
        Ek_Gn=(n*c_star/0.5124)**2-V0        # Ek at G
        Ek_Zn=((n+0.5)*c_star/0.5124)**2-V0    # Ek at Z
        hv_Gn=round(Ek_Gn+work_fct,2)
        hv_Zn=round(Ek_Zn+work_fct,2)
        print("\n G["+str(n)+"]:  hv = Ek + Phi = "+str(round(hv_Gn,2))+" eV ")
        print(" Z["+str(n)+"]:  hv = Ek + Phi = "+str(round(hv_Zn,2))+" eV")

def Print_Gamma_hv(lattice,V0,hv1,hv2):
    work_fct=EA.wk
    c_star=2*np.pi/lattice                # 2pi/c = GZ distance
    Ek1=hv1-work_fct
    Ek2=hv2-work_fct
    k_z1=0.5124*np.sqrt(Ek1+V0)    # Kz at (kx,ky)=0 i.e. cos(th)=1
    k_z2=0.5124*np.sqrt(Ek2+V0)    # Kz at (kx,ky)=0 i.e. cos(th)=1
    GZ_n1=round((k_z1/c_star*2),1)
    G_n1=round((k_z1/c_star),1)
    GZ_n2=round((k_z2/c_star*2),1)
    G_n2=round((k_z2/c_star),1)
    if modf(G_n1)[0]>=0.5:
        n1=modf(G_n1)[1]+1,0
    else:
        n1=modf(G_n1)[1]
    n2=modf(G_n2)[1]
    print("\n hv1 = "+str(hv1)+" eV:  " +str(GZ_n1)+" * pi/c = " +str(G_n1)+" * 2pi/c")
    if type(n1) == tuple: n1 = n1[0]
    if type(n2) == tuple: n2 = n2[0]
    Print_Gamma_n(lattice,V0,n1,n2)
    print("\n hv2 = "+str(hv2)+" eV:  " +str(GZ_n2)+" * pi/c = " +str(G_n2)+" * 2pi/c")
    return n1,n2

def kx2deg(lattice,hv):
    a=np.pi/lattice
    b=0.5124*np.sqrt(hv)
    c=a/b
    theta_rad=np.asin(c)
    theta_deg=rad2deg(theta_rad)
    print(" 1/2-BZ (GX) = "+str(round(theta_deg,1))+" deg")
    
def resolution_EA(PE,slit_SES):    # updated 10/30/17: straight slits scaled to slit width not area
    SES_Table={}
    SES_Table[2]   = {1: 1.6, 2:2,   3:2,   4:4,   5:4,   6:6,    7:12,   8:20,   9:32}
    SES_Table[5]   = {1: 2.7, 2:4,   3:4,   4:7,   5:7,   6:11,   7:20,   8:34,   9:54}
    SES_Table[10]  = {1: 5.7, 2:9,   3:9,   4:14,  5:14,  6:23,   7:43,   8:71,   9:114}
    SES_Table[20]  = {1:10.8, 2:16,  3:16,  4:27,  5:27,  6:43,   7:81,   8:135,  9:216}
    SES_Table[50]  = {1:34.6, 2:52,  3:52,  4:87,  5:87,  6:138,  7:260,  8:433,  9:692}
    SES_Table[100] = {1:49.5, 2:74,  3:74,  4:124, 5:124, 6:198,  7:371,  8:619,  9:990}
    SES_Table[200] = {1:88.9, 2:133, 3:133, 4:222, 5:222, 6:356,  7:667,  8:1111, 9:1778}
    SES_Table[500] = {1:250,  2:375, 3:375, 4:625, 5:625, 6:1000, 7:1875, 8:3125, 9:5000}
    try:
        SES=SES_Table[PE][slit_SES]
    except KeyError:
        print("WARNING: Not a valid PE/Slit combination")
        SES=0
    return SES


def SES_slit_set(val):
    """
    Set the Scienta Slit
    val    w_mm    l_mm    Shape
    1    0.2        25    Curved
    2    0.3        30    Straight
    3    0.3        25    Curved
    4    0.5        30    Straight
    5    0.5        25    Curved
    6    0.8        25    Curved
    7    1.5        30    Straight
    8    2.5        30    Straight
    8.5            open apperture
    9    4.0        30    Straight    
    """
    caput("29idc:m8.VAL",val,wait=True,timeout=18000)


def SES_slit_get():
    SES=caget("29idc:m8.RBV")
    return SES
